//
//  agent.h
//  Infektsimulation
//
//  Created on 19.08.13.
//  Alexander Droste & Florian Flachsenberg
//

#include <stdbool.h>
#include "datatypes.h"
#include <mpi.h>
#include <omp.h>

#ifndef Infektsimulation_agent_h 
#define Infektsimulation_agent_h

//radii of infection paths
#define radius_air 5
#define radius_skin 2
#define radius_oral 2
#define radius_xxx 1

typedef struct{
    //probabilities for transmission types
    Percent air;    //respiration
    Percent skin;   //touch
    Percent oral;   //oral
    Percent xxx;    //sexual
}InfectionProb;

typedef struct {
	Percent mortality; /* percentage of mortality */
	Percent immunogenicity; /* percentage of immunogenicity */
	InfectionProb infection_prob; /* probability of infection */
	Percent outbreak_prob; /* probability infection leads to disease */
    Percent mutation_prob; /* probability of mutation on infection */

	unsigned short incubation_mean_dur; /* mean duration of incubation */
	unsigned short incubation_max_dur; /* incubation max duration */
	unsigned short illness_mean_dur; /* mean duration of illness */
	unsigned short illness_max_dur; /* max duration of illness */
	char infectious_incubation; /* infectious during incubation time (0,1) */

    unsigned short mutation_counter;
    unsigned short mutation_counter_threshold;
    unsigned long immunogenicity_factor; /* in case more than one agent is around they can be identified */
} Agent;

//MAL = malicious
typedef enum {Influenza, Ebola, Herpes, Smallpox} AgentType;

Agent agent_new_stereotype(AgentType agent_type);
//Agent* agent_copy(Agent* inc_agent);
//void agent_delete(Agent* inc_agent);
void agent_print_properties(Agent* inc_agent);
void agent_mutate(Agent* agent);
bool agent_is_valid(Agent agent);

#endif
