//
//  datatypes.c
//  Infektsimulation
//
//  Created on 21.08.13.
//  Alexander Droste & Florian Flachsenberg
//

#include <stdio.h>
#include "datatypes.h"
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <assert.h>
#include "world.h"

#define INVALID_MATRIX_CREATION 700


//––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
//private methods
static bool isMatrixValid (FieldMatrix* matrix);
//––––––––––––Field Matrix––––––––––––––––––––––––––––––––––––––––––––––––––

FieldMatrix *createMatrix(const unsigned int rows, const unsigned int cols, unsigned f_a_length)
{
    extern World world;
	FieldMatrix *matrix = malloc(sizeof(FieldMatrix));
	matrix->m = rows; matrix->n = cols;
    
	matrix->comp = calloc(rows, sizeof(Field*));
	matrix->data = calloc((rows*cols), sizeof(Field));

    int i,j;
    
	if (isMatrixValid(matrix)) {
		for (i = 0; i < rows; ++i) {
			//dem reihenvektor werden spaltenvektoren zugewiesen
			matrix->comp[i] = &(matrix->data[i * matrix->n]);
            for (j = 0; j < cols; ++j) {
                matrix->comp[i][j].healthy = list_new();
                matrix->comp[i][j].infected = list_new();
                //max y index fields might have smaller heigth
                if (i == (rows-1)) {
                    matrix->comp[i][j].size.height = world.size.height - (f_a_length*i);
                }
                //the others have full height
                else{
                    matrix->comp[i][j].size.height = f_a_length;
                }
                //max x index fields might have smaller width
                if (j == (cols-1)) {
                    matrix->comp[i][j].size.width = world.size.width - (f_a_length*j);
                }
                //the others have full width
                else{
                    matrix->comp[i][j].size.width = f_a_length;
                }
                
            }
		}
	}
	else {
        printf("\nerror: invalid matrix creation");
        exit(INVALID_MATRIX_CREATION);
	}
    
	return matrix;
}

Field* get_field_from_matrix_for_coord(unsigned int x, unsigned int y)
{
    extern World world;
    //[row][col] => [y][x] is correct
    return &(world.field_matrix->comp[y][x]);
}

void freeMatrix(FieldMatrix *matrix)
{
    extern World world;
    Field* f;
    //free lists
    for (int x = 0; x < world.field_matrix->n ; ++x) {
        for (int y = 0; y < world.field_matrix->m; ++y) {
            f = get_field_from_matrix_for_coord(x, y);
            list_delete(f->healthy);
            list_delete(f->infected);
        }
    }
    
	if (matrix->data) {
		free(matrix->data);
	}
	if (matrix->comp) {
		free(matrix->comp);
	}
	if (matrix) {
		free(matrix);
	}
}

static bool isMatrixValid (FieldMatrix* matrix)
{
    bool valid; //set 0 if not valid
    if(NULL == (matrix->data) || NULL == matrix->comp){
        valid = 0;
    }
    else{
        valid = 1;
    }
    return valid;
}
//––––––––––––Linked List–––––––––––––––––––––––––––––––––––––––––––––––––
void* list_element_get_value(const List_element* el)
{
    #if DEBUG
        assert(el);
    #endif
    return el->data;
}

List_element* list_first(const List* l)
{
    #if DEBUG
        assert(l);
    #endif
    return l->first;
}

List_element* list_last(const List* l)
{
    #if DEBUG
        assert(l);
    #endif

    return l->last;
}

List* list_new(void)
{
    List* l = malloc(sizeof(List));
    #if DEBUG
        assert(l);
    #endif
    l->first = l->last = NULL;
    return l;
}

List_element* list_element_next(const List_element* le)
{
    #if DEBUG
        assert(le);
    #endif
    return le->next;
}

List_element* list_element_prev(const List_element* le)
{
    #if DEBUG
        assert(le);
    #endif
    return le->prev;
}

List_element* list_append(List* l, void* value)
{
    List_element* new_el = malloc(sizeof(List_element));
    #if DEBUG
        assert(new_el);
    #endif

    new_el->next = NULL;
    new_el->data = value;
    new_el->prev = l->last;
    if (l->last)
        l->last->next = new_el;
    else
        l->first = new_el;
    l->last = new_el;
    
    #if DEBUG
        assert(new_el != new_el->prev);
    #endif
    
    return new_el;
}

void list_delete_element(List* l, List_element* el)
{
    #if DEBUG
        assert(l);
        assert(el);
    #endif
    
    if (el->prev != NULL)
        el->prev->next = el->next;
    else
        l->first = el->next;
    if (el->next != NULL)
        el->next->prev = el->prev;
    else
        l->last = el->prev;
    
    free(el);
    el = NULL;
}

void list_transfer_element(List* dest, List* src, List_element* el)
{
    #if DEBUG
        assert(dest);
        assert(src);
        assert(el);
        assert(dest!=src);
    #endif
    
    if (el->prev != NULL)
        el->prev->next = el->next;
    else
        src->first = el->next;
    
    if (el->next != NULL)
        el->next->prev = el->prev;
    else
        src->last = el->prev;

    el->next = NULL;
    el->prev = dest->last;
    if (dest->last)
        dest->last->next = el;
    else
        dest->first = el;
    dest->last = el;
    
    #if DEBUG
        assert(el != el->prev);
    #endif
}

void list_delete(List* l)
{
    #if DEBUG
        assert(l);
    #endif
    
    List_element* curr = l->first;
    List_element* next;
    while(curr != NULL){
        next = curr->next;
        free(curr);
        curr = next;
    }
    free(l);
    l = NULL;
}

/**
 *  Iterate over list with arbitrary function that returns an int
 *  and has void* as its only argument.
 *  !At the moment not in use!
 *
 *  @param l list to iterate over
 *  @param function this function used on all objects in list
 *
 *  @return    the return value can be used to count things
 *              or evaluate a boolean expression for a list.
 */
int list_iterate_over_all_elements_with_function(List* l, int (*function)(void*))
{
    //safe init with 0
    int x = 0;
    
    //iterate over all elements in both lists
    List_element* element = list_first(l);
    while (element) {
        void* element_data = element->data;
        
        x += function(element_data);
        list_element_next(element);
    }
    return x;
}
//––––––––––––Dynamic Array––––––––––––––––––––––––––––––––––––––––––––––––––
DynArray array_create()
{
    DynArray array;
    array.memory = NULL;
    array.allocated = 0; //measured in bytes
    array.used = 0;      //measured in bytes
    array.max_index = -1;
    return array;
}

void array_add(DynArray *array, unsigned short data)
{
    size_t size = sizeof(unsigned short);
    if ((array->allocated - array->used) < size) {
       
        //make space for 2 more elements
        size_t to_allocate = (size*2 + array->allocated);
        
        array->memory = realloc(array->memory, to_allocate);
        array->allocated = to_allocate;
    }
    
    ++array->max_index;
    array->memory[array->max_index] = data;
    array->used = array->used + size;
    
    #if DEBUG
        assert(data == array->memory[array->max_index]);
    #endif

}

void array_delete(DynArray *array)
{
    free(array->memory);
    //no malloc was called do not free!
    //free(array); 
}

int array_length(DynArray *array)
{
    return array->max_index + 1;
}

unsigned short array_get(DynArray *array, int index)
{
    if (index < 0 || index > array->max_index)
        return 0;
    
    return array->memory[index];
}

void array_insert(DynArray *array, int index, unsigned short data)
{
    if (index < 0 || index > array->max_index)
        return;
    
    array->memory[index] = data;
}

bool array_is_element_contained(DynArray *array, unsigned short data)
{
    bool is_contained = false;
    for(int i = 0; i < array_length(array); ++i)
    {
//        printf("%i\n", array->memory[i]);
        if (data == array->memory[i]) {
            is_contained = true;
        }
    }
    return is_contained;
}

//––––––––––––Time Functions–––––––––––––––––––––––––––––––––––––––––––––––––
int time_minutes_to_seconds(double minutes)
{
    return (minutes * 60);
}
int time_hours_to_seconds(double hours)
{
    return (hours * 3600);
}
int time_days_to_seconds(double days)
{
    return (days * 86400);
}
int time_weeks_to_seconds(double weeks)
{
    return (weeks * 604800);
}
int time_months_to_seconds(double months)
{
    //ref: google
    //https://encrypted.google.com/search?q=seconds%20per%20day#fp=f775caf3a08584&q=seconds+per+month
    return (months * 2629743);
}
int time_years_to_seconds(double years)
{
    //ref: google
    //https://encrypted.google.com/search?q=seconds%20per%20day#fp=f775caf3a08584&q=seconds+per+year
    return (years * 31556926);
}

//––––––––––––Helper Functions–––––––––––––––––––––––––––––––––––––––––––––––––
/**
 *  Euclidean square dist (without root calc, because of speed)
 *
 *  @param x1 x1
 *  @param y1 y1
 *  @param x2 x2
 *  @param y2 y2
 *
 *  @return distance
 */
unsigned sq_euclidean (unsigned x1, unsigned y1, unsigned x2, unsigned y2)
{
  return (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1);
}

/**
 *  create random 1 or -1
 *
 *  @return randon sign
 */
very_short_int rnd_sign(gsl_rng* rng)
{
    very_short_int sign = rnd_int(rng) % 2;
    if (sign == 0) {
        sign = -1;
    }
    return sign;
}

double rnd_one_percent(gsl_rng* rng)
{
    return gsl_rng_uniform(rng);
}

double rnd_percent(gsl_rng* rng)
{
    return gsl_rng_uniform(rng)*100;
}

unsigned long int rnd_int(gsl_rng* rng)
{
    return gsl_rng_get(rng);
}

double keep_in_range(double min, double max, double inc_value)
{
    double checked_value;
    //value too big
    if (inc_value > max) {
        checked_value = max;
    }
    //value too small
    else if (inc_value < min){
        checked_value = min;
    }
    //value was in range
    else{
        checked_value = inc_value;
    }
    return checked_value;
}
