//
//  datatypes.h
//  Infektsimulation
//
//  Created on 20.08.13.
//  Alexander Droste & Florian Flachsenberg
//

#include <stdbool.h>
#include <gsl_rng.h>


#ifndef Infektsimulation_datatypes_h
#define Infektsimulation_datatypes_h


typedef  float Percent;
typedef char very_short_int;

typedef struct{
    unsigned int x;
    unsigned int y;
}Position;

typedef struct{
    unsigned int x;
    unsigned int y;
}Coordinates;

typedef struct{
    unsigned int width;
    unsigned int height;
}Size;

//––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
//marcos
//choose the larger of two values
#define max(a,b)            \
({ __typeof__ (a) _a = (a); \
__typeof__ (b) _b = (b);    \
_a > _b ? _a : _b; })

//––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
typedef struct List_element List_element;
typedef struct List List;
typedef struct FieldMatrix FieldMatrix;
typedef struct Field Field;

//––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
struct Field
{
    List* healthy;
    List* infected;
    unsigned nof_infected, nof_healthy;
    Size size;
};

struct FieldMatrix
{
    unsigned int m, n; // Rows, cols
    Field** comp;  // Pointers to rows in data
    Field* data; // Data, ordered by row, then by col
    char* name;
};

typedef struct{
    int rows;
    int cols;
}MatrixDimensions;

typedef struct
{
    unsigned short *memory;
    size_t allocated;
    size_t used;
    int max_index;
}DynArray;

struct List_element {
    struct List_element* prev;
    struct List_element* next;
    void* data;
};

struct List {
    List_element* first;
    List_element* last;
};
//––––––––––––Field Matrix–––––––––––––––––––––––––––––––––––––––––––––––––
FieldMatrix *createMatrix(const unsigned int rows, const unsigned int cols, unsigned f_a_length);
void freeMatrix(FieldMatrix *matrix);
Field* get_field_from_matrix_for_coord(unsigned int x, unsigned int y);

//––––––––––––Linked List–––––––––––––––––––––––––––––––––––––––––––––––––
void* list_element_get_value(const List_element *element); /* read value from element */
List* list_new(void);/* create new double linked list */
List_element* list_first(const List* list); /* first list element */
List_element* list_last(const List* list);/* last list element */
List_element* list_element_next(const List_element* element); /* successor element */
List_element* list_element_prev(const List_element* element); /* predeccessor element */
List_element* list_append(List* l, void* value);  /* append new element */
void list_delete_element (List* list, List_element* element); /* delete element */
void list_transfer_element (List* dest, List* src, List_element* element); /* transfer element between lists */
void list_delete(List* list); /* delete list with all elements*/

//executes the given function on the !data of all elements
int list_iterate_over_all_elements_with_function(List* l, int (*function)(void*));

//––––––––––––Dynamic Array––––––––––––––––––––––––––––––––––––––––––––––––––
DynArray array_create(); //returns no pointer!
void array_add(DynArray *array, unsigned short data);
void array_delete(DynArray *array);
int array_length(DynArray *array);
bool array_is_element_contained(DynArray *array, unsigned short data);
unsigned short array_get(DynArray *array, int index);
//––––––––––––Time Functions–––––––––––––––––––––––––––––––––––––––––––––––––
int time_minutes_to_seconds(double minutes);
int time_hours_to_seconds(double hours);
int time_days_to_seconds(double days);
int time_weeks_to_seconds(double weeks);
int time_months_to_seconds(double months);
int time_years_to_seconds(double years);

//––––––––––––Helper Functions–––––––––––––––––––––––––––––––––––––––––––––––––
unsigned sq_euclidean(unsigned x1, unsigned y1, unsigned x2, unsigned y2);
unsigned long int rnd_int(gsl_rng* rng);
double rnd_one_percent(gsl_rng* rng);
double rnd_percent(gsl_rng* rng);
very_short_int rnd_sign(gsl_rng* rng);
double keep_in_range(double min, double max, double inc_value);

#endif
