//
//  host.h
//  Infektsimulation
//
//  Created on 19.08.13.
//  Alexander Droste & Florian Flachsenberg
//

#include "agent.h"
#include <stdbool.h>
#include <mpi.h>
#include <omp.h>

#ifndef Infektsimulation_host_h //include guard
#define Infektsimulation_host_h


typedef enum {HEALTHY, INFECTED, ILL, IMMUNE, DEAD} HostState;

typedef struct {
	Agent agent;
	int state;
    
	unsigned short outbreak_time; /* start point of disease */
    unsigned short infection_time; /* time of infection */
    
	unsigned short end_of_disease; /* end point of disease */
	char will_ill; /* host will get disease after infection */
	char will_die; /* host will die on end of disease */
	char will_immune; /* host will get immune on end of disease */
    
    unsigned long immunities; /* agents host is immune against */
    unsigned short disappear_duration; /*time until host disappears after death*/
    
    Position pos; /* host position: x/y value */
    Field* field; /* field the host is positioned on */

    List_element* list_element; /* pointer to list element so hosts can be deleted from lists easily */
    
} Host;

typedef struct {
  Host host;
  Host* phost; // pointer to Host (CAUTION: only valid in "home" process)
  Coordinates field_coord[4];
  // save for each possible neighbour field the coordinates) and
  unsigned curr_field_coord;
} HostForInfection;

Host* host_create();
void host_move(Host*);
void host_process(Host* host, unsigned short inc_time);
void host_prepare_infection_process(Host *host, unsigned short time);
void host_start_disease(Host *host);
void host_end_disease(Host *host);
void host_process(Host *host, unsigned short inc_time);
void host_infect(Host *host, Agent inc_agent, unsigned inc_time);
void host_infect_sim_start(Host *host, Agent inc_agent);
void host_delete(Host* host);
//void host_delete_agent(Host* host);
void host_for_infection_process(HostForInfection* host_for_infection, unsigned time);

#endif
