//
// mpi_datatypes.c
// Infektsimulation
//
// Created on 08.09.13.
// Alexander Droste & Florian Flachsenberg
//

#include <mpi.h>
#include "host.h"
#include "datatypes.h"
#include "agent.h"

// non-extern declaraction of global MPI_Datatypes here
#define DATATYPE_INIT
#include "mpi_datatypes.h"

void initialize_mpi_datatypes()
{
  // mpi_infectionProb
  MPI_Aint air_address, skin_address, oral_address,
      xxx_address, address;
  InfectionProb ip;

  MPI_Get_address(&ip, &address);
  MPI_Get_address(&ip.air, &air_address);
  MPI_Get_address(&ip.oral, &oral_address);
  MPI_Get_address(&ip.skin, &skin_address);
  MPI_Get_address(&ip.xxx, &xxx_address);

  int ip_blocks[] = {1,1,1,1};
  MPI_Aint ip_offset[4];
  MPI_Datatype ip_types[4] = {MPI_PERCENT, MPI_PERCENT, MPI_PERCENT, MPI_PERCENT};

  ip_offset[0] = air_address - address;
  ip_offset[1] = skin_address - address;
  ip_offset[2] = oral_address - address;
  ip_offset[3] = xxx_address - address;

  MPI_Type_create_struct(4, ip_blocks, ip_offset, ip_types, &mpi_infectionProb);
  MPI_Type_commit(&mpi_infectionProb);

  // mpi_agent
  MPI_Aint mortality_address, immunogenicity_address, infection_prob_address,
      outbreak_prob_address, mutation_prob_address, incubation_mean_dur_address,
      incubation_max_dur_address, illness_mean_dur_address, illness_max_dur_address,
      infectious_incubation_address, mutation_counter_address,
      mutation_counter_threshold_address, immunogenicity_factor_address;
  Agent agent;

  MPI_Get_address(&agent, &address);
  MPI_Get_address(&agent.mortality, &mortality_address);
  MPI_Get_address(&agent.immunogenicity, &immunogenicity_address);
  MPI_Get_address(&agent.infection_prob, &infection_prob_address);
  MPI_Get_address(&agent.outbreak_prob, &outbreak_prob_address);
  MPI_Get_address(&agent.mutation_prob, &mutation_prob_address);
  MPI_Get_address(&agent.incubation_mean_dur, &incubation_mean_dur_address);
  MPI_Get_address(&agent.incubation_max_dur, &incubation_max_dur_address);
  MPI_Get_address(&agent.illness_mean_dur, &illness_mean_dur_address);
  MPI_Get_address(&agent.illness_max_dur, &illness_max_dur_address);
  MPI_Get_address(&agent.infectious_incubation, &infectious_incubation_address);
  MPI_Get_address(&agent.mutation_counter, &mutation_counter_address);
  MPI_Get_address(&agent.mutation_counter_threshold, &mutation_counter_threshold_address);
  MPI_Get_address(&agent.immunogenicity_factor, &immunogenicity_factor_address);

  int agent_blocks[] = {1,1,1,1,1,1,1,1,1,1,1,1,1};
  MPI_Aint agent_offset[13];
  MPI_Datatype agent_types[13] = {MPI_PERCENT, MPI_PERCENT, mpi_infectionProb, 
      MPI_PERCENT, MPI_PERCENT, MPI_UNSIGNED_SHORT, MPI_UNSIGNED_SHORT,
      MPI_UNSIGNED_SHORT, MPI_UNSIGNED_SHORT, MPI_CHAR, MPI_UNSIGNED_SHORT,
      MPI_UNSIGNED_SHORT, MPI_UNSIGNED_LONG};
  //TODO: MPI_UNSIGNED_LONG is 32bit on 32bit machines!

  agent_offset[0] = mortality_address - address;
  agent_offset[1] = immunogenicity_address - address;
  agent_offset[2] = infection_prob_address - address;
  agent_offset[3] = outbreak_prob_address - address;
  agent_offset[4] = mutation_prob_address - address;
  agent_offset[5] = incubation_mean_dur_address - address;
  agent_offset[6] = incubation_max_dur_address - address;
  agent_offset[7] = illness_mean_dur_address - address;
  agent_offset[8] = illness_max_dur_address - address;
  agent_offset[9] = infectious_incubation_address - address;
  agent_offset[10] = mutation_counter_address - address;
  agent_offset[11] = mutation_counter_threshold_address - address;
  agent_offset[12] = immunogenicity_factor_address - address;

  MPI_Type_create_struct(13, agent_blocks, agent_offset, agent_types, &mpi_agent);
  MPI_Type_commit(&mpi_agent);

  // mpi_position
  MPI_Aint x_address, y_address;
  Position position;

  MPI_Get_address(&position, &address);
  MPI_Get_address(&position.x, &x_address);
  MPI_Get_address(&position.y, &y_address);

  int position_blocks[] = {1,1};
  MPI_Aint position_offset[2];
  MPI_Datatype position_types[2] = {MPI_UNSIGNED, MPI_UNSIGNED};

  position_offset[0] = x_address - address;
  position_offset[1] = y_address - address;

  MPI_Type_create_struct(2, position_blocks, position_offset, 
      position_types, &mpi_position);
  MPI_Type_commit(&mpi_position);

  // mpi_coordinates
  MPI_Aint x_address2, y_address2;
  Coordinates coord;

  MPI_Get_address(&coord, &address);
  MPI_Get_address(&coord.x, &x_address2);
  MPI_Get_address(&coord.y, &y_address2);

  int coord_blocks[] = {1,1};
  MPI_Aint coord_offset[2];
  MPI_Datatype coord_types[2] = {MPI_UNSIGNED, MPI_UNSIGNED};

  coord_offset[0] = x_address2 - address;
  coord_offset[1] = y_address2 - address;

  MPI_Type_create_struct(2, coord_blocks, coord_offset, 
      coord_types, &mpi_coordinates);
  MPI_Type_commit(&mpi_coordinates);


  //mpi_host
  MPI_Aint agent_address, state_address, outbreak_time_address,
           infection_time_address, end_of_disease_address,
           will_ill_address, will_die_address, will_immune_address,
           immunities_address, disappear_duration_address,
           pos_address, field_address, list_element_address;
  Host host;

  MPI_Get_address(&host, &address);
  MPI_Get_address(&host.agent, &agent_address);
  MPI_Get_address(&host.state, &state_address);
  MPI_Get_address(&host.outbreak_time, &outbreak_time_address);
  MPI_Get_address(&host.infection_time, &infection_time_address);
  MPI_Get_address(&host.end_of_disease, &end_of_disease_address);
  MPI_Get_address(&host.will_ill, &will_ill_address);
  MPI_Get_address(&host.will_die, &will_die_address);
  MPI_Get_address(&host.will_immune, &will_immune_address);
  MPI_Get_address(&host.immunities, &immunities_address);
  MPI_Get_address(&host.disappear_duration, &disappear_duration_address);
  MPI_Get_address(&host.pos, &pos_address);
  MPI_Get_address(&host.field, &field_address);
  MPI_Get_address(&host.list_element, &list_element_address);


  int host_blocks[] = {1,1,1,1,1,1,1,1,1,1,1,1,1};
  MPI_Aint host_offset[13];
  MPI_Datatype host_types[13] = {mpi_agent, MPI_INT,
      MPI_UNSIGNED_SHORT, MPI_UNSIGNED_SHORT, MPI_UNSIGNED_SHORT,
      //TODO: here problem, MPI_UNSIGNED_LONG different on 32/64bit
      MPI_CHAR, MPI_CHAR, MPI_CHAR, MPI_UNSIGNED_LONG,
      MPI_UNSIGNED_SHORT, mpi_position,
      //and here MPI_UNSIGNED_LONG is fine! => sizeof of pointer
      MPI_UNSIGNED_LONG, MPI_UNSIGNED_LONG};

  host_offset[0] = agent_address - address;
  host_offset[1] = state_address - address;
  host_offset[2] = outbreak_time_address - address;
  host_offset[3] = infection_time_address - address;
  host_offset[4] = end_of_disease_address - address;
  host_offset[5] = will_ill_address - address;
  host_offset[6] = will_die_address - address;
  host_offset[7] = will_immune_address - address;
  host_offset[8] = immunities_address - address;
  host_offset[9] = disappear_duration_address - address;
  host_offset[10] = pos_address - address;
  host_offset[11] = field_address - address;
  host_offset[12] = list_element_address - address;

  MPI_Type_create_struct(13, host_blocks, host_offset, 
      host_types, &mpi_host);
  MPI_Type_commit(&mpi_host);

  // mpi_position
  MPI_Aint host_address, phost_address, fields_address, currfield_address;
  HostForInfection hostFI;

  MPI_Get_address(&hostFI, &address);
  MPI_Get_address(&hostFI.host, &host_address);
  MPI_Get_address(&hostFI.field_coord, &fields_address);
  MPI_Get_address(&hostFI.curr_field_coord, &currfield_address);
  MPI_Get_address(&hostFI.phost, &phost_address);

  int hostfi_blocks[] = {1,4,1,1};
  MPI_Aint hostfi_offset[4];
  MPI_Datatype hostfi_types[4] = {mpi_host, mpi_coordinates, MPI_UNSIGNED, MPI_UNSIGNED_LONG};

  hostfi_offset[0] = host_address - address;
  hostfi_offset[1] = fields_address - address;
  hostfi_offset[2] = currfield_address - address;
  hostfi_offset[3] = phost_address - address;

  MPI_Type_create_struct(4, hostfi_blocks, hostfi_offset, 
      hostfi_types, &mpi_hostForInfection);
  MPI_Type_commit(&mpi_hostForInfection);


//  printf("initialized mpi datatypes...\n");
}
