//
// test_mpi_cust_types.c
// Infektsimulation
//
// Created on 09.09.13.
// Alexander Droste & Florian Flachsenberg
//

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include "host.h"
#include "agent.h"
#include "mpi_datatypes.h"

int main(int argc, char** argv)
{
    MPI_Init(&argc, &argv);
    //size = number of processes
    int size, rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if (size < 2) {
      printf("error, at least to workers required!\n");
      return 1;
    }
    else if (size > 2) {
      printf("warning, all workers >= 2 ignored!\n");
    }

    initialize_mpi_datatypes();
    
    InfectionProb ip1;
    ip1.air = 1.0;
    ip1.skin = 2.0;
    ip1.oral = 3.0;
    ip1.xxx = 4.0;

    Agent agent1;
    agent1.mortality = 5.0;
    agent1.immunogenicity = 6.0;
    agent1.infection_prob = ip1;
    agent1.outbreak_prob = 7.0;
    agent1.mutation_prob = 8.0;
    agent1.incubation_mean_dur = 9;
    agent1.incubation_max_dur = 10;
    agent1.illness_mean_dur = 11;
    agent1.illness_max_dur = 12;
    agent1.infectious_incubation = 13;
    agent1.mutation_counter = 14;
    agent1.mutation_counter_threshold = 15;
    agent1.immunogenicity_factor = 16L;

    Position position1;
    position1.x = 17;
    position1.y = 18;

    Host host1;
    host1.agent = agent1;
    host1.state = 19;
    host1.outbreak_time = 20;
    host1.infection_time = 21;
    host1.end_of_disease = 22;
    host1.will_ill = 23;
    host1.will_die = 24;
    host1.will_immune = 25;
    host1.immunities = 26L;
    host1.disappear_duration = 27;
    host1.pos = position1;
    host1.field = (Field*)28L;
    host1.list_element = (List_element*)29L;

    HostForInfection hostFI1;
    hostFI1.host = host1;
    hostFI1.phost = (Host*)30L;
    hostFI1.field_coord[0].x = 1;
    hostFI1.field_coord[0].y = 2;
    hostFI1.field_coord[1].x = 3;
    hostFI1.field_coord[1].y = 4;
    hostFI1.field_coord[2].x = 5;
    hostFI1.field_coord[2].y = 6;
    hostFI1.field_coord[3].x = 7;
    hostFI1.field_coord[3].y = 8;
    hostFI1.curr_field_coord = 10;


    if (rank == 0) {
      MPI_Send(&ip1, 1, mpi_infectionProb, 1, 0, MPI_COMM_WORLD);
      MPI_Send(&agent1, 1, mpi_agent, 1, 1, MPI_COMM_WORLD);
      MPI_Send(&position1, 1, mpi_position, 1, 2, MPI_COMM_WORLD);
      MPI_Send(&host1, 1, mpi_host, 1, 3, MPI_COMM_WORLD);
      MPI_Send(&hostFI1, 1, mpi_hostForInfection, 1, 4, MPI_COMM_WORLD);
    }
    else if (rank == 1) {
      InfectionProb ip2;
      Agent agent2;
      Position position2;
      Host host2;
      HostForInfection hostFI2;

      MPI_Recv(&ip2, 1, mpi_infectionProb, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Recv(&agent2, 1, mpi_agent, 0, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Recv(&position2, 1, mpi_position, 0, 2, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Recv(&host2, 1, mpi_host, 0, 3, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Recv(&hostFI2, 1, mpi_hostForInfection, 0, 4, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

      // introduce test error
      // ip2.air += 0.000001;

      double diff = 0.0;
      diff += fabs(ip1.air - ip2.air);
      diff += fabs(ip1.skin - ip2.skin);
      diff += fabs(ip1.oral -ip2.oral);
      diff += fabs(ip1.xxx -ip2.xxx);
      printf("Infection probability: %.12lf\n", diff);

      diff = 0.0;
      diff += fabs(agent1.mortality - agent2.mortality);
      diff += fabs(agent1.immunogenicity - agent2.immunogenicity);
      diff += fabs(agent1.infection_prob.air - agent2.infection_prob.air);
      diff += fabs(agent1.infection_prob.skin - agent2.infection_prob.skin);
      diff += fabs(agent1.infection_prob.oral - agent2.infection_prob.oral);
      diff += fabs(agent1.infection_prob.xxx - agent2.infection_prob.xxx);
      diff += fabs(agent1.outbreak_prob - agent2.outbreak_prob);
      diff += fabs(agent1.mutation_prob - agent2.mutation_prob);
      diff += fabs((double)agent1.incubation_mean_dur - agent2.incubation_mean_dur);
      diff += fabs((double)agent1.incubation_max_dur - agent2.incubation_max_dur);
      diff += fabs((double)agent1.illness_mean_dur - agent2.illness_mean_dur);
      diff += fabs((double)agent1.illness_max_dur - agent2.illness_max_dur);
      diff += fabs((double)agent1.infectious_incubation - agent2.infectious_incubation);
      diff += fabs((double)agent1.mutation_counter - agent2.mutation_counter);
      diff += fabs((double)agent1.mutation_counter_threshold - agent2.mutation_counter_threshold);
      diff += fabs((double)agent1.immunogenicity_factor - agent2.immunogenicity_factor);
      printf("Agent: %.12lf\n", diff);

      diff = 0.0;
      diff += fabs((double)position1.x - position2.x);
      diff += fabs((double)position1.y - position2.y);
      printf("Position: %.12lf\n", diff);

      // introduce test error
      //host2.list_element = 1L;

      diff = 0.0;
      diff += fabs(host1.agent.mortality - host2.agent.mortality);
      diff += fabs(host1.agent.immunogenicity - host2.agent.immunogenicity);
      diff += fabs(host1.agent.infection_prob.air - host2.agent.infection_prob.air);
      diff += fabs(host1.agent.infection_prob.skin - host2.agent.infection_prob.skin);
      diff += fabs(host1.agent.infection_prob.oral - host2.agent.infection_prob.oral);
      diff += fabs(host1.agent.infection_prob.xxx - host2.agent.infection_prob.xxx);
      diff += fabs(host1.agent.outbreak_prob - host2.agent.outbreak_prob);
      diff += fabs(host1.agent.mutation_prob - host2.agent.mutation_prob);
      diff += fabs((double)host1.agent.incubation_mean_dur - host2.agent.incubation_mean_dur);
      diff += fabs((double)host1.agent.incubation_max_dur - host2.agent.incubation_max_dur);
      diff += fabs((double)host1.agent.illness_mean_dur - host2.agent.illness_mean_dur);
      diff += fabs((double)host1.agent.illness_max_dur - host2.agent.illness_max_dur);
      diff += fabs((double)host1.agent.infectious_incubation - host2.agent.infectious_incubation);
      diff += fabs((double)host1.agent.mutation_counter - host2.agent.mutation_counter);
      diff += fabs((double)host1.agent.mutation_counter_threshold - host2.agent.mutation_counter_threshold);
      diff += fabs((double)host1.agent.immunogenicity_factor - host2.agent.immunogenicity_factor);
      diff += ((double)host1.state - host2.state);
      diff += ((double)host1.outbreak_time - host2.outbreak_time);
      diff += ((double)host1.infection_time - host2.infection_time);
      diff += ((double)host1.end_of_disease -host2.end_of_disease);
      diff += ((double)host1.will_ill - host2.will_ill);
      diff += ((double)host1.will_die - host2.will_die);
      diff += ((double)host1.will_immune - host2.will_immune);
      diff += ((double)host1.immunities - host2.immunities);
      diff += ((double)host1.disappear_duration - host2.disappear_duration);
      diff += fabs((double)host1.pos.x - host2.pos.x);
      diff += fabs((double)host1.pos.y - host2.pos.y);
      diff += ((unsigned long)host1.field - (unsigned long)host2.field);
      diff += ((unsigned long)host1.list_element - (unsigned long)host2.list_element);
      printf("Host: %.12lf\n", diff);

      diff = 0.0;
      diff += fabs(hostFI1.host.agent.mortality - hostFI2.host.agent.mortality);
      diff += fabs(hostFI1.host.agent.immunogenicity - hostFI2.host.agent.immunogenicity);
      diff += fabs(hostFI1.host.agent.infection_prob.air - hostFI2.host.agent.infection_prob.air);
      diff += fabs(hostFI1.host.agent.infection_prob.skin - hostFI2.host.agent.infection_prob.skin);
      diff += fabs(hostFI1.host.agent.infection_prob.oral - hostFI2.host.agent.infection_prob.oral);
      diff += fabs(hostFI1.host.agent.infection_prob.xxx - hostFI2.host.agent.infection_prob.xxx);
      diff += fabs(hostFI1.host.agent.outbreak_prob - hostFI2.host.agent.outbreak_prob);
      diff += fabs(hostFI1.host.agent.mutation_prob - hostFI2.host.agent.mutation_prob);
      diff += fabs((double)hostFI1.host.agent.incubation_mean_dur - hostFI2.host.agent.incubation_mean_dur);
      diff += fabs((double)hostFI1.host.agent.incubation_max_dur - hostFI2.host.agent.incubation_max_dur);
      diff += fabs((double)hostFI1.host.agent.illness_mean_dur - hostFI2.host.agent.illness_mean_dur);
      diff += fabs((double)hostFI1.host.agent.illness_max_dur - hostFI2.host.agent.illness_max_dur);
      diff += fabs((double)hostFI1.host.agent.infectious_incubation - hostFI2.host.agent.infectious_incubation);
      diff += fabs((double)hostFI1.host.agent.mutation_counter - hostFI2.host.agent.mutation_counter);
      diff += fabs((double)hostFI1.host.agent.mutation_counter_threshold - hostFI2.host.agent.mutation_counter_threshold);
      diff += fabs((double)hostFI1.host.agent.immunogenicity_factor - hostFI2.host.agent.immunogenicity_factor);
      diff += ((double)hostFI1.host.state - hostFI2.host.state);
      diff += ((double)hostFI1.host.outbreak_time - hostFI2.host.outbreak_time);
      diff += ((double)hostFI1.host.infection_time - hostFI2.host.infection_time);
      diff += ((double)hostFI1.host.end_of_disease -hostFI2.host.end_of_disease);
      diff += ((double)hostFI1.host.will_ill - hostFI2.host.will_ill);
      diff += ((double)hostFI1.host.will_die - hostFI2.host.will_die);
      diff += ((double)hostFI1.host.will_immune - hostFI2.host.will_immune);
      diff += ((double)hostFI1.host.immunities - hostFI2.host.immunities);
      diff += ((double)hostFI1.host.disappear_duration - hostFI2.host.disappear_duration);
      diff += fabs((double)hostFI1.host.pos.x - hostFI2.host.pos.x);
      diff += fabs((double)hostFI1.host.pos.y - hostFI2.host.pos.y);
      diff += ((unsigned long)hostFI1.host.field - (unsigned long)hostFI2.host.field);
      diff += ((unsigned long)hostFI1.host.list_element - (unsigned long)hostFI2.host.list_element);
      diff += fabs((unsigned long)hostFI1.phost - (unsigned long)hostFI2.phost);
      diff += fabs((double)hostFI1.field_coord[0].x - hostFI2.field_coord[0].x);
      diff += fabs((double)hostFI1.field_coord[0].y - hostFI2.field_coord[0].y);
      diff += fabs((double)hostFI1.field_coord[1].x - hostFI2.field_coord[1].x);
      diff += fabs((double)hostFI1.field_coord[1].y - hostFI2.field_coord[1].y);
      diff += fabs((double)hostFI1.field_coord[2].x - hostFI2.field_coord[2].x);
      diff += fabs((double)hostFI1.field_coord[2].y - hostFI2.field_coord[2].y);
      diff += fabs((double)hostFI1.field_coord[3].x - hostFI2.field_coord[3].x);
      diff += fabs((double)hostFI1.field_coord[3].y - hostFI2.field_coord[3].y);
      diff += fabs((double)hostFI1.curr_field_coord - hostFI2.curr_field_coord);
      printf("infection struct: %.12f\n", diff);
    }

    MPI_Finalize();
    return 0;
}
