//
// test_mpi_poll.c
// Infektsimulation
//
//
// Created on 10.09.13.
// Alexander Droste & Florian Flachsenberg
//

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <stdbool.h>
#include <unistd.h>
#include "agent.h"
#include <gsl/gsl_rng.h> 
#include "datatypes.h"
#include "host.h"
#include "world.h"
#include "mpi_datatypes.h"

void test_polling(){
    
    int np, me;
    int tag = 42;
    MPI_Comm_size(MPI_COMM_WORLD, &np);
    MPI_Comm_rank(MPI_COMM_WORLD, &me);
    
    //    Host* host_array[10000];
    Host** host_array = malloc(sizeof(Host*)*10000);
    for (int i=0; i<10000; ++i) {
        host_array[i] = calloc(1, sizeof(Host));
        host_array[i]->outbreak_time = i;
    }
    
    if (me == 0) {
        MPI_Request r1;
        for (int i = 0; i < 10000; ++i) {
            
            MPI_Isend(host_array[i], sizeof(Host), MPI_BYTE, 1, tag, MPI_COMM_WORLD, &r1);
//            MPI_Send(host_array[i], sizeof(Host), MPI_BYTE, 1, tag, MPI_COMM_WORLD);
//            MPI_Issend(host_array[i], sizeof(Host), MPI_BYTE, 1, tag, MPI_COMM_WORLD, &r1);
//            MPI_Rsend(host_array[i], sizeof(Host), MPI_BYTE, 1, tag, MPI_COMM_WORLD);
            MPI_Wait(&r1, MPI_STATUS_IGNORE);
            //save to call free directly
            free(host_array[i]);
        }
    }
    
    if (me == 1){
        MPI_Request r2;
        Host recv_buff;
        
        int rec_message = 0;
        int message_count = 0;
        int cont = 1;
        int check_for_message = 1;
        long prev = - 1;
        bool receivedonehosttwice = 0;
        
        while (cont) {
            
            if (check_for_message) {
                MPI_Irecv(&recv_buff, sizeof(Host), MPI_BYTE, 0, tag, MPI_COMM_WORLD, &r2);
                check_for_message = 0;
            }
            
            MPI_Test(&r2, &rec_message, MPI_STATUS_IGNORE);
            
            if (rec_message) {
                
                if (prev == recv_buff.outbreak_time || !(recv_buff.outbreak_time > prev)) {
                    receivedonehosttwice = 1;
                    printf("received one host twice");
                    
                }
                prev = recv_buff.outbreak_time;
                printf ("busy: Process %d received value %i\n", me, recv_buff.outbreak_time);
                check_for_message = 1;
                ++message_count;
                
                if (message_count == 10000) {
                    cont = 0;
                }
            }
        }
        //number of correct runs
        static int corr = 0;
        if (message_count == 10000 && !receivedonehosttwice) {
            ++corr;
            printf("\n!receive succesful! %i correct rounds\n", corr);
        }
    }
}

int main(int argc, char* argv[]) {

    MPI_Init(NULL, NULL);
	//test i rounds 
    for (int i = 0; i < 5; ++i) {
        test_polling();
    }

    MPI_Finalize();
    return 0;
}
