//
//  world.h
//  Infektsimulation
//
//  Created on 21.08.13.
//  Alexander Droste & Florian Flachsenberg
//

#include <limits.h>
#include "host.h"
#include "agent.h"
#include <stdbool.h>
#include "datatypes.h"

#ifndef Infektsimulation_world_h
#define Infektsimulation_world_h

typedef struct{
    Size size; //complete size of the world
    FieldMatrix* field_matrix; //matrix containing all fields
    Size full_size_one_field; //size of one field
    unsigned infected_total, immune_total, died_total, ill_total, ill_current, healthy_current, infected_current, infectious_dead_current;
    unsigned min_incubation, max_incubation, min_illness, max_illness;
    double sum_incubation, sum_illness;
    unsigned radius_air_sq, radius_oral_sq, radius_skin_sq, radius_xxx_sq, maxradius_sq;
    unsigned height_process;
    unsigned width_process;
    unsigned width_process_mid;
    unsigned height_process_mid;
    unsigned number_of_cols;
    unsigned number_of_cols_mid;
    unsigned number_of_rows;
    unsigned number_of_rows_mid;
    unsigned hosts_send_count, hosts_received_count; //number of hosts send
    unsigned infects_send_count, infects_received_count; //number of hosts send
    unsigned residents;
    MatrixDimensions process_matrix_dimensions; //how the processes are arranged in the world
    unsigned number_of_neighbours;
}World;

Coordinates world_get_rel_coord_for_host(Host* inc_host);
Field* world_get_field_for_host(Host* inc_host);
void world_transfer_host_to_field(Host* inc_host, Field* dest_field);
Position world_get_pos_within_field(Host* host);
void world_add_host_to_field(Host* inc_host);
void world_remove_host_from_field(Host* host);
unsigned world_get_process_for_host(Host* inc_host);
bool world_host_position_collides(Host* inc_host);
void world_create(unsigned int world_axes_length, unsigned int field_axes_length);
void world_create_for_agent(unsigned world_axes_length, Agent agent);
void world_destroy();
bool world_host_position_collides(Host* inc_host);
void world_simulate(unsigned short time_steps, unsigned stop_if_finished);
Coordinates world_get_abs_for_rel_coord(unsigned x, unsigned y);
unsigned world_get_process_for_absolute_coord(unsigned x, unsigned y);
Coordinates world_get_rel_for_abs_coord(unsigned x, unsigned y);

#endif
